import os
from abc import ABC, abstractmethod
from typing import List

from amsatop._htop.process import Process


class Htop(ABC):
    """
    Abstract base class representing a simplified version of the Unix 'htop' utility.

    This class defines the interface for interacting with system process information.
    It is designed to be subclassed by students or developers who will provide
    concrete implementations for the defined methods.

    Attributes:
        proc_folder (str): Path to the directory containing process information.
            Defaults to "/proc", but can be overridden by setting the "PROC_FOLDER"
            environment variable. You should *allways* use this variable and not "/proc".
    """

    proc_folder: str

    def __init__(self, proc_folder: str = os.getenv("PROC_FOLDER", "/proc")) -> None:
        """
        Initialize the Htop base class with a specified process folder.

        :param proc_folder: Path to the proc filesystem used to read process information.
                           If not provided, it defaults to the value of the "PROC_FOLDER"
                           environment variable, or "/proc" if the variable is unset.
        """
        super().__init__()
        self.proc_folder = proc_folder

    @abstractmethod
    def get_processes(self) -> List[Process]:
        """
        Retrieve all system processes, without the priority field

        :return: A list of Process objects representing all running processes.
        :raises NotImplementedError: If not implemented by subclass.
        """
        raise NotImplementedError("Prac-2.1")

    @abstractmethod
    def get_priorities(self) -> List[Process]:
        """
        Retrieve all system processes along with their scheduling priorities.

        :return: A list of Process objects with priority information.
        :raises: NotImplementedError: If not implemented by subclass.
        """
        raise NotImplementedError("Prac-2.2")

    @abstractmethod
    def get_hup(self) -> List[Process]:
        """
        Retrieve processes that are eligible to ignore the SIGHUP signal.

        :return: A list of Process objects that ignore the SIGHUP signal.
        :raises: NotImplementedError: If not implemented by subclass.
        """
        raise NotImplementedError("Prac-2.3")



from dataclasses import dataclass
from enum import Enum


class Type(Enum):
    """
    Enumeration of process types.
    """

    TASK = "task"
    THREAD = "thread"
    KTHREAD = "kthread"


@dataclass(frozen=True)
class Process:
    """
    Immutable data class representing a system process.

    :param pid: The unique process ID.
    :type pid: int

    :param command: The command or executable that launched the process.
    :type command: str

    :param type: The type of the process
    :type type: Type

    :param priority: The scheduling priority of the process.
                     Can be None if unavailable or you're doing Prac-2.1 .
    :type priority: int | None
    """

    pid: int
    command: str
    type: Type
    priority: int | None



"""
This module provides utility logic to simplify the implementation of amsatop.
While it's possible to build your solution without it, these components are designed to make your life easier.
"""

from amsatop.utils.__stat_file import StatFile, get_stat_file_from_path
from amsatop.utils.__status_file import StatusFile, get_status_file_from_path

__all__ = ["StatusFile", "StatFile", "get_status_file_from_path", "get_stat_file_from_path"]


"""
amsatop package.

Exposes top-level components for creating your own htop
"""

from amsatop._htop.htop import Htop
from amsatop._htop.htop_mock import HtopMock
from amsatop._htop.process import Process, Type
from amsatop._ui import run_ui
import amsatop.utils

__all__ = ["Htop", "HtopMock", "Process", "Type", "run_ui", "utils"]



import logging
from dataclasses import dataclass, field
from typing import List


@dataclass(frozen=True)
class StatusFile:
    """
    Represents the parsed contents of a `/proc/[pid or tgid]/status` file on a Linux system.
    All fields map to their corresponding values in the status file.
    """
    Name: str
    State: str
    Tgid: int
    Ngid: int
    Pid: int
    PPid: int
    TracerPid: int
    Uid: List[int]
    Gid: List[int]
    FDSize: int
    Groups: List[int] = field(default_factory=list)
    NStgid: int = 0
    NSpid: int = 0
    NSpgid: int = 0
    NSsid: int = 0
    Kthread: int = 0
    VmPeak: int = 0
    VmSize: int = 0
    VmLck: int = 0
    VmPin: int = 0
    VmHWM: int = 0
    VmRSS: int = 0
    RssAnon: int = 0
    RssFile: int = 0
    RssShmem: int = 0
    VmData: int = 0
    VmStk: int = 0
    VmExe: int = 0
    VmLib: int = 0
    VmPTE: int = 0
    VmSwap: int = 0
    HugetlbPages: int = 0
    CoreDumping: int = 0
    THP_enabled: int = 0
    untag_mask: str = ''
    Threads: int = 0
    SigQ: str = ''
    SigPnd: str = ''
    ShdPnd: str = ''
    SigBlk: str = ''
    SigIgn: str = ''
    SigCgt: str = ''
    CapInh: str = ''
    CapPrm: str = ''
    CapEff: str = ''
    CapBnd: str = ''
    CapAmb: str = ''
    NoNewPrivs: int = 0
    Seccomp: int = 0
    Seccomp_filters: int = 0
    Speculation_Store_Bypass: str = ''
    SpeculationIndirectBranch: str = ''
    Cpus_allowed: str = ''
    Cpus_allowed_list: str = ''
    Mems_allowed: str = ''
    Mems_allowed_list: str = ''
    voluntary_ctxt_switches: int = 0
    nonvoluntary_ctxt_switches: int = 0
    x86_Thread_features: str | None = ''
    x86_Thread_features_locked: str | None = ''
    Umask: str | None = None

    @classmethod
    def __from_lines(cls, lines: List[str]) -> "StatusFile":
        def parse_value(key: str, value: str):
            value = value.strip()
            if not value:
                return "" if "x86_Thread" in key else []

            if key in {"Uid", "Gid", "Groups"}:
                return list(map(int, value.split()))

            if key.startswith("Vm") or key in {
                "HugetlbPages", "voluntary_ctxt_switches", "nonvoluntary_ctxt_switches"
            }:
                return int(value.replace("kB", "").strip())

            if key in {
                "Tgid", "Ngid", "Pid", "PPid", "TracerPid", "FDSize", "NStgid",
                "NSpid", "NSpgid", "NSsid", "Kthread", "CoreDumping", "THP_enabled",
                "Threads", "NoNewPrivs", "Seccomp", "Seccomp_filters"
            }:
                return int(value)

            return value

        data = {}
        for line in lines:
            if not line.strip() or ':' not in line or line.strip() == "EOF":
                continue
            key, value = line.split(":", 1)
            field_key = key.strip().replace("-", "_")
            parsed_value = parse_value(key.strip(), value)
            data[field_key] = parsed_value

        return cls(**data)


def get_status_file_from_path(path: str) -> StatusFile | None:
    """
    Given a path, returns a StatusFile, which can be useful for analyzing a process or thread.

    :param path: path to the status file (will be usually in the form of /proc/[pid or tgid]/status
    :returns: StatusFile or None if the path doesn't exist (or other unhandled exception)
    :raises: ValueError if the path isn't valid.
    """
    if not path.endswith("/status"):
        raise ValueError("Status file, should end with /status.....")

    try:
        with open(path, "r", encoding="utf-8") as f:
            return StatusFile.__from_lines(f.readlines())
    except FileNotFoundError:
        return None
    except Exception as e:
        logging.error(
            f"Unexpected error in amsatop library (not caused by this code). "
            f"Please report it to your instructors. Error: {e}"
        )
        return None



from dataclasses import dataclass
from enum import Enum
from typing import List, Literal

from amsatop._htop.htop import Htop
from amsatop._htop.process import Process


class ViewMode(Enum):
    PROCESSES = "processes"
    PRIORITIES = "priorities"
    HUP = "hup"

    @staticmethod
    def from_str(s: str) -> "ViewMode":
        for mode in ViewMode:
            if s == mode.value:
                return mode
        raise ValueError(f"Invalid ViewMode, can't convert it from string: {s}")


@dataclass
class UiState:
    __htop: Htop
    __mode: ViewMode = ViewMode.PROCESSES

    def change_view(
        self, mode: Literal["processes", "priorities", "hup"] | str
    ) -> None:
        self.__mode = ViewMode.from_str(mode)

    @property
    def processes(self) -> List[Process]:
        match self.__mode:
            case ViewMode.PROCESSES:
                return self.__htop.get_processes()
            case ViewMode.PRIORITIES:
                return self.__htop.get_priorities()
            case ViewMode.HUP:
                return self.__htop.get_hup()
            case _:
                raise NotImplementedError()



import logging
from dataclasses import dataclass


@dataclass(frozen=True)
class StatFile:
    """
    Represents key fields from a process's or thread /proc/[pid or tgid]/stat file.
    """
    pid: int
    command: str        # Filename of the executable (in parentheses)
    state: str          # Process state
    ppid: int           # Parent PID
    pgrp: int           # Process group ID
    session: int        # Session ID
    tty_nr: int         # Controlling terminal
    tpgid: int          # Foreground process group ID
    flags: int          # Kernel flags
    minflt: int         # Minor faults
    cminflt: int        # Minor faults with children
    majflt: int         # Major faults
    cmajflt: int        # Major faults with children
    utime: int          # User mode jiffies
    stime: int          # Kernel mode jiffies
    cutime: int         # User mode jiffies with children
    cstime: int         # Kernel mode jiffies with children
    priority: int       # Priority value
    nice: int           # Nice value
    num_threads: int    # Number of threads
    starttime: int      # Start time since boot (in jiffies)

    @classmethod
    def __from_stat_line(cls, line: str) -> "StatFile":
        first_paren = line.find('(')
        last_paren = line.rfind(')')
        if first_paren == -1 or last_paren == -1:
            raise ValueError("Malformed stat line: missing parentheses around command")
        before = line[:first_paren].strip()
        command = line[first_paren + 1:last_paren]
        after = line[last_paren + 1:].strip()
        fields = before.split() + [command] + after.split()
        if len(fields) < 22:
            raise ValueError("Malformed stat line: insufficient fields")

        return StatFile(
            pid=int(fields[0]),
            command=fields[1],
            state=fields[2],
            ppid=int(fields[3]),
            pgrp=int(fields[4]),
            session=int(fields[5]),
            tty_nr=int(fields[6]),
            tpgid=int(fields[7]),
            flags=int(fields[8]),
            minflt=int(fields[9]),
            cminflt=int(fields[10]),
            majflt=int(fields[11]),
            cmajflt=int(fields[12]),
            utime=int(fields[13]),
            stime=int(fields[14]),
            cutime=int(fields[15]),
            cstime=int(fields[16]),
            priority=int(fields[17]),
            nice=int(fields[18]),
            num_threads=int(fields[19]),
            starttime=int(fields[21]),
        )


def get_stat_file_from_path(path: str) -> StatFile | None:
    """
    Given a path, returns a StatFile, which can be usefull for analyzing a process or thread.

    :param path: path to the stat file (will be usually in the form of /proc/[pid or tgid]/stat
    :returns: StatFile or None if the path doesn't exist.
    :raises: ValueError if the path isn't valid.
    """
    if not path.endswith("/stat"):
        raise ValueError(f"Stat file, should end with /stat......")

    try:
        with open(path, "r", encoding="utf-8") as f:
            line = f.readline()
            return StatFile.__from_stat_line(line)
    except FileNotFoundError:
        return None
    except Exception as e:
        logging.error(
            "Unhandled error in amsatop library (not your solution), please report to the instructors: %s",
            e
        )
        return None


import pyfiglet
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import DataTable, Footer, Header, Static

from amsatop._htop.htop import Htop
from amsatop._ui.state import UiState


class HtopTUI(App[None]):
    BINDINGS = [
        Binding("p", "toggle_view('processes')", "Processes", show=True),
        Binding("r", "toggle_view('priorities')", "Priorities", show=True),
        Binding("h", "toggle_view('hup')", "Ignored hup", show=True),
        Binding("q", "quit", "Quit"),
    ]

    def __init__(self, htop: Htop, refresh_seconds: int):
        super().__init__()
        self.state = UiState(htop)
        self.refresh_seconds = refresh_seconds

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static(pyfiglet.figlet_format("AMSA25-26"), id="ascii")
        yield DataTable(id="process_table")
        yield Footer()

    def on_mount(self) -> None:
        table = self.query_one(selector=DataTable)
        table.add_columns("PID", "Command", "Type", "Priority")
        self.update_table()
        self.set_interval(interval=self.refresh_seconds, callback=self.update_table)

    def update_table(self) -> None:
        table = self.query_one(DataTable)
        table.clear()
        for process in self.state.processes:
            priority = str(process.priority) if process.priority is not None else "N/A"
            table.add_row(
                str(process.pid), process.command, process.type.value, priority
            )

    def action_toggle_view(self, view: str) -> None:
        self.state.change_view(mode=view)
        self.update_table()



import random
from typing import override

from amsatop._htop.htop import Htop
from amsatop._htop.process import Process, Type


class HtopMock(Htop):
    """
    Mock implementation of the Htop interface for testing or demonstration purposes.

    This class simulates process information by returning hardcoded `Process` instances.
    Each method (`get_processes`, `get_priorities`, `get_hup`) returns a random subset
    of predefined mock processes.

    Useful for testing code that depends on Htop without requiring access to the actual
    system process tree.
    """

    RETURN_PROCESSES = 5

    @override
    def get_processes(self) -> list[Process]:
        not_real_processes = [
            Process(pid=1, command="init", type=Type.TASK, priority=None),
            Process(pid=2, command="bash", type=Type.THREAD, priority=None),
            Process(pid=3, command="kworker", type=Type.KTHREAD, priority=None),
            Process(pid=4, command="python", type=Type.TASK, priority=None),
            Process(pid=5, command="top", type=Type.THREAD, priority=None),
            Process(pid=6, command="_htop", type=Type.TASK, priority=None),
            Process(pid=7, command="systemd", type=Type.KTHREAD, priority=None),
            Process(pid=8, command="tmux", type=Type.THREAD, priority=None),
            Process(pid=9, command="node", type=Type.TASK, priority=None),
            Process(pid=10, command="java", type=Type.KTHREAD, priority=None),
            Process(pid=11, command="emacs", type=Type.THREAD, priority=None),
            Process(pid=12, command="vim", type=Type.TASK, priority=None),
            Process(pid=13, command="zsh", type=Type.THREAD, priority=None),
            Process(pid=14, command="docker", type=Type.KTHREAD, priority=None),
        ]
        return random.sample(
            not_real_processes, min(self.RETURN_PROCESSES, len(not_real_processes))
        )

    @override
    def get_priorities(self) -> list[Process]:
        not_real_processes = [
            Process(pid=15, command="sshd", type=Type.TASK, priority=10),
            Process(pid=16, command="nginx", type=Type.THREAD, priority=5),
            Process(pid=17, command="postgres", type=Type.KTHREAD, priority=1),
            Process(pid=18, command="redis", type=Type.TASK, priority=7),
            Process(pid=19, command="chrome", type=Type.THREAD, priority=8),
            Process(pid=20, command="firefox", type=Type.TASK, priority=6),
            Process(pid=21, command="docker", type=Type.KTHREAD, priority=3),
            Process(pid=22, command="mysql", type=Type.TASK, priority=9),
            Process(pid=23, command="emacs", type=Type.THREAD, priority=4),
            Process(pid=24, command="vim", type=Type.TASK, priority=2),
            Process(pid=25, command="zsh", type=Type.THREAD, priority=5),
            Process(pid=26, command="python", type=Type.TASK, priority=7),
            Process(pid=27, command="java", type=Type.KTHREAD, priority=6),
        ]
        return random.sample(
            not_real_processes, min(self.RETURN_PROCESSES, len(not_real_processes))
        )

    @override
    def get_hup(self) -> list[Process]:
        not_real_processes = [
            Process(pid=28, command="sshd", type=Type.TASK, priority=10),
            Process(pid=29, command="nginx", type=Type.THREAD, priority=5),
            Process(pid=30, command="postgres", type=Type.KTHREAD, priority=1),
            Process(pid=31, command="redis", type=Type.TASK, priority=7),
            Process(pid=32, command="chrome", type=Type.THREAD, priority=8),
            Process(pid=33, command="firefox", type=Type.TASK, priority=6),
            Process(pid=34, command="docker", type=Type.KTHREAD, priority=3),
            Process(pid=35, command="mysql", type=Type.TASK, priority=9),
            Process(pid=36, command="emacs", type=Type.THREAD, priority=4),
            Process(pid=37, command="vim", type=Type.TASK, priority=2),
            Process(pid=38, command="zsh", type=Type.THREAD, priority=5),
            Process(pid=39, command="python", type=Type.TASK, priority=7),
            Process(pid=40, command="java", type=Type.KTHREAD, priority=6),
        ]
        return random.sample(
            not_real_processes, min(self.RETURN_PROCESSES, len(not_real_processes))
        )



from amsatop._htop.htop import Htop
from amsatop._htop.htop_mock import HtopMock
from amsatop._ui.tui import HtopTUI


def run_ui(htop: Htop = None, refresh: int = 2) -> None:
    """
    Given an instance of htop, run the tui interface.
    By default, if no parameter is given, it launches the HtopMock implementation.

    :param htop: An instance of your htop implementation
    :param refresh: The seconds between each table refresh, 2 by default
    """
    if htop is None:
        htop = HtopMock()
    app = HtopTUI(htop=htop, refresh_seconds=refresh)
    app.run()


__all__ = ["run_ui"]



"""
This module contains all the logic that you need for implementing amsatop
"""

from amsatop._htop.htop import Htop
from amsatop._htop.htop_mock import HtopMock
from amsatop._htop.process import Process, Type

__all__ = ["Htop", "HtopMock", "Process", "Type"]
